#define globalDefine
#include <FbdReader.hxx>

#define PLOT 0
 
using namespace std;

//The shell sewing tolerance is used to evaluate which faces
//are neighbouring faces and have to be connected
//If neighbouring faces are not connected, decrease tolerance
const double shellSewingTolerance = 0.0000001;

bool hashing;

//Calculate point intex from point name
//Works only for names that where automatically generated by cgx (Hash Code)
int pointGetIndex(const char * name)
{
  int charValue,hashValue,index=0;
  if(strlen(name)<=4)//Valid point name
  {
    for(int i=0;i<=3;i++)
    {
      charValue=(int)name[i];
      if(charValue >= 48 && charValue <= 57)//number 0-9
      {
	hashValue=charValue-48;//hash values: 0-9 --> 0-9 
      }
      else if(charValue >= 65 && charValue <= 90)//letter A-Z
      {
	hashValue=(charValue-65)+10;//hash values: A-Z --> 10-36 	
      }
      else
      {
	cout<<"invalid character"<<endl;
	return 0;
      }	
      if(i==0) hashValue-=13; //first point is D001
      index+=pow(36,3-i)*hashValue;
    }
    return index;
  }
  else
  {
    cout<<name<<" is no valid point name -> no hashing"<<endl;
    return 0;
  }
}

bool parseFBD(char * filename)
{
    string line;      
    input.open(filename);
    if(input.is_open())
    {
      while(getline(input,line))//Parse through input file linewise
      {
	//preparing the input text
	for(int i=0;line[i]!=0;i++)
	{
	  if(line[i]<=122 && line[i]>=97)
	    line[i]-=32;//Make line uppercase (Works only for ASCII Symbols)
	  if(line[i]==33 || line[i]==37)//Remove ! and % characters from aliasing
	    line[i]=32;//Replace them with blanks
	}
	
	//****************************************//
	//******* Search for Points: PNT *********//
	//****************************************//
	if(line.find("PNT")<=3)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[30];
	  //Proceede to next word
	  strLine>>cBuff;
	  strLine>>cBuff;
	  //Save Name
	  Point * tmpPnt = new Point(cBuff);
	  if(hashing){
	    for(int i=pointVec.size(); i<pointGetIndex(cBuff)-1;i++)
	    {
	      pointVec.push_back(new Point("dummy"));
	    }
	  }
	  //Proceede to next word 
	  strLine>>cBuff;	
	  tmpPnt->setX(atof(cBuff));	  	  	  	  
	  //Proceede to next word
	  strLine>>cBuff;	
	  tmpPnt->setY(atof(cBuff));
	  //Proceede to next word
	  strLine>>cBuff;	
	  tmpPnt->setZ(atof(cBuff));
	  pointVec.push_back(tmpPnt);
	}
	
	//********************************************//
	//***** Search for sequential sets: SEQUA ****//
	//********************************************//
	if(line.find("SEQA")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16],name[16];
	  
	  //Proceede to next word
	  strLine>>cBuff;
	  strLine>>cBuff;
	  //Save Name
	  strcpy(name,cBuff); 
	  //Proceede to next word
	  strLine>>cBuff;  
	  Seqa * tmpSeq = NULL;
	  if(strcmp(cBuff,"PNT")==0)//New Seqa
	  {
	    tmpSeq = new Seqa(name);    
	    seqaVec.push_back(tmpSeq);
	  }
	  else if(strcmp(cBuff,"END")==0)//Add to Seqa
	  {
	    for(vector<Seqa*>::iterator s = seqaVec.begin();s != seqaVec.end();++s)
	    {	    
		if(strcmp(name,(*s)->getName())==0){
		  tmpSeq = (*s); break;}
	    }
	  }
	  while(strLine>>cBuff && tmpSeq != NULL)
	  {    
	    if(hashing){
	      if(strcmp(cBuff,"PNT")!=0)
		tmpSeq->addPoint(pointVec[pointGetIndex(cBuff)-1]);	      
	    }
	    else{
	      for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	      {
		if(strcmp(cBuff,(*p)->getName())==0){
		  tmpSeq->addPoint(*p);
		  //cout<<cBuff<<" vs "<<pointVec[pointGetIndex(cBuff)-1]->getName()<<endl;
		  break;
		}  
	      }
	    }
	  }  
	}
	
	//****************************************************//  	
	//Search for 2D Elements Lines, Arcs, Splines: LINE **//
	//****************************************************//
	if(line.find("LINE")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16],name[16];
	  Point * p1, *p2, *cnt = NULL;
	  Seqa * tmpSeq = NULL;
	  //Proceede to next word
	  strLine>>cBuff;
	  strLine>>cBuff;
	  strcpy(name,cBuff);	  
	  //Porceede to next word
	  strLine>>cBuff; //Contains now name of P1
	  //Search for p1
	  for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	  {
	    if(strcmp(cBuff,(*p)->getName())==0){
	      p1=*p; break;}	 
	  }
	  	 
	  strLine>>cBuff; //Contains now name of P2
	  //Search for p2
	  
	  for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	  {
	    if(strcmp(cBuff,(*p)->getName())==0){
	      p2=*p; break;}  
	  }
	  
	  strLine>>cBuff; //Contains now the third argument
	  //Search for a Center Point
	  for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	  {
	    if(strcmp(cBuff,(*p)->getName())==0){
	      cnt = *p; 
	      break;
	    }	  
	  }
	  //Search for a Sequential Set
	  if(cnt == NULL){
	    for(vector<Seqa*>::iterator s = seqaVec.begin();s != seqaVec.end();++s)
	    {	    
		if(strcmp(cBuff,(*s)->getName())==0){
		  tmpSeq = *s; 
		  break;
		}
	    }	    
	  }	  
	  char  cDiv[4];
	  int div;
	  if(tmpSeq != NULL)//Generate Spline
	  {	    	 
	      Spline * tmpSpli = new Spline(name);    	     
	      elem2dVec.push_back(tmpSpli);
	      tmpSpli->setSeqa(tmpSeq);
	      //Proceede to next word	         
	      strLine>>cBuff;
	      strcpy(cDiv,cBuff);      
	      if(!(strLine>>cBuff)) cDiv[0]=' ';//div has one or two digits -> erase first digit
	      div = atoi(cDiv);//get div
	      //cout<<"Div: "<<div<<endl;
	      tmpSpli->setDiv(div);
	      tmpSpli->makeEdge();
	  }
	  else if(cnt != NULL)//Generate Arc
	  {
	      Arc * tmpArc = new Arc(name);
	      elem2dVec.push_back(tmpArc);
	      tmpArc->setP1(p1);
	      tmpArc->setP2(p2);
	      tmpArc->setCnt(cnt);
	      //Proceede to next word
	      strLine>>cBuff;
	      strcpy(cDiv,cBuff);      
	      if(!(strLine>>cBuff)) cDiv[0]=' ';//div has one or two digits -> erase first digit
	      div = atoi(cDiv);//get div
	      //cout<<"Div: "<<div<<endl;
	      tmpArc->setDiv(div);
	      tmpArc->makeEdge();
	  }
	  else//Generate Line
	  {
	      Line * tmpLine = new Line(name);
	      elem2dVec.push_back(tmpLine);
	      tmpLine->setP1(p1);
	      tmpLine->setP2(p2);
	      tmpLine->makeEdge();
	      strcpy(cDiv,cBuff);   
	      if(!(strLine>>cBuff)) cDiv[0]=' ';//div has one or two digits -> erase first digit
	      div = atoi(cDiv);//get div
	      tmpLine->setDiv(div);
	      //cout<<"LINE "<<name<<" Div: "<<div<<endl;
	  }
	}
	
	//*********************************************//  	
	//******* Search for NURBS Lines: NURL ********//
	//*********************************************//
	if(line.find("NURL")<=4)
	{
          stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16];	  
	  BSplineCurve * tmpNurbsL;
	  Point * p1, *p2;
	  bool end = false, found,compact = false, multi = false;
	  strLine>>cBuff;//Proceede to next word
	  strLine>>cBuff;//Proceede to next word	    
	  tmpNurbsL = new BSplineCurve(cBuff);
	  strLine>>cBuff;//Proceede to next word
	  int deg,nKnt=0,nPol;	  
	  strLine>>cBuff;//Contains now p1 or COMPACT
	  if(strcmp(cBuff,"COMPACT")==0)
	  {
	    compact=true,
	    strLine>>cBuff;//Contains now p1
	  }
	  
	  for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	  {
	    if(strcmp(cBuff,(*p)->getName())==0){
	      p1=*p; break;}	 
	  }	  
	  
	  strLine>>cBuff;//Contains now p2	
	  for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	  {
	    if(strcmp(cBuff,(*p)->getName())==0){
	      p2=*p; break;}  
	  }	  
	  
	  strLine>>cBuff;//Contains now deg
	  deg=atoi(cBuff);
	  
	  strLine>>cBuff;//Contains now nPoles	
	  nPol=atoi(cBuff);
          
          strLine>>cBuff;//Contains now nKnots	

	  TColgp_Array1OfPnt poles(1,nPol); 
	  gp_Pnt pnt1(*(p1->getX()),*(p1->getY()),*(p1->getZ()));	
	  gp_Pnt pnt2(*(p2->getX()),*(p2->getY()),*(p2->getZ()));
	  poles.SetValue(1,pnt1);
	  poles.SetValue(nPol,pnt2);
	  
	  TColStd_Array1OfReal weights(1,nPol); 

	  vector<double> ktmp;
	  
	  vector<int> mtmp;
	  
	  strLine>>cBuff;//Proceede to next Line: DEFINE, CONTROL, KNOT or END
	  int initial=0, save=0;
          do{    
	    //Save Line as vector of strings for easier access 
	    getline(input,line);
	    //Make line uppercase (Works only for ASCII Symbols)
	    for(int i=0;line[i]!=0;i++)
	    {
	      if(line[i]<=122 && line[i]>=97)
		line[i]-=32;
	    }
	    stringstream tmp;
	    tmp<<line;
	    vector<string> content;	    
	    while (tmp >> cBuff) {
	      content.push_back(cBuff);
	    }
	 
	    initial = save;//Save the position where the last control point was found
	    
            if(strcmp(content[2].c_str(),"CONTROL")==0 )
	    {
	      if(compact)
	      {
		if(atoi(content[3].c_str()) != 1 && atoi(content[3].c_str()) != nPol)
		{
		  gp_Pnt pnt(atof(content[4].c_str()),atof(content[5].c_str()),atof(content[6].c_str()));	
		  poles.SetValue(atoi(content[3].c_str()),pnt);
		}
		weights.SetValue(atoi(content[3].c_str()),atof(content[7].c_str()));
	      }
	      else
	      {
		if(atoi(content[3].c_str()) != 1 && atoi(content[3].c_str()) != nPol)
		{
		  found = false;
		  //Starting the search from 'initial' was implemented to speed up the search in case of Control Points written in consecutive lines
		  do{
		    for(vector<Point*>::iterator p = pointVec.begin()+initial;p != pointVec.end();++p)
		    {
		      
		      if(strcmp(content[4].c_str(),(*p)->getName())==0){
			gp_Pnt pnt(*((*p)->getX()),*((*p)->getY()),*((*p)->getZ()));
			poles.SetValue(atoi(content[3].c_str()),pnt);
			found = true;
			break;
		      }  
		      save++;
		    }
		    if (!found) save = 0;		  
		    initial = 0;//In case the Control Point was not written in a consecutive line, start searching again from the beginning
		  }while(found == false);
		}
		weights.SetValue(atoi(content[3].c_str()),atof(content[5].c_str()));
	      }
	    }
	    else if(strcmp(content[2].c_str(),"KNOT")==0)
	    {
		if(atoi(content[3].c_str())>1){
		  //cout<<atof(content[5].c_str())<<" =? "<<*(uKtmp.end()-1)<<endl;
		    if(atof(content[4].c_str())==*(ktmp.end()-1))
		      multi = true;
		    else 
		      multi = false;
		}
		else{
		  multi = false;
		}
		
		if(multi)//Not a new knot -> Increase Multiplicity
		{
		  (*(mtmp.end()-1))++;
		}
		else//New knot
		{
		  ktmp.push_back(atof(content[4].c_str()));
		  mtmp.push_back(1);
		  nKnt++;
		} 	
	    }
	    else if(strcmp(content[2].c_str(),"END")==0)
	    {  
	      TColStd_Array1OfReal knots(1,nKnt);   
	      TColStd_Array1OfInteger mults(1,nKnt);   
	      int c=1;
	      for(vector<double>::iterator e = ktmp.begin();e != ktmp.end();++e)
	      {
		knots.SetValue(c,*e);
		mults.SetValue(c,mtmp[c-1]);
		c++;
	      }  

	      Handle(Geom_BSplineCurve) aCurve = new Geom_BSplineCurve(poles,weights,knots,mults,deg);
              
	   #if PLOT == 1
              cout<<"NURBS Line: "<<tmpNurbsL->getName()<<endl;
              cout<<"Deg: "<<aCurve->Degree()<<endl; 
	      for(int i=1;i<=aCurve->NbPoles();i++)
              {
                cout<<"Poles: "<< (aCurve->Pole(i)).X()<<" "<<(aCurve->Pole(i)).Y()<<" "<< (aCurve->Pole(i)).Z()<<" Weights: "<<aCurve->Weight(i)<<endl ;
              }
              
              for(c=1;c<=aCurve->NbKnots();c++)
	      {
		cout<<"Knots "<<aCurve->Knot(c)<<" Mults: "<<aCurve->Multiplicity(c)<<endl;
	      }
            #endif  
	      tmpNurbsL->setCurve(aCurve);
	      elem2dVec.push_back(tmpNurbsL);
	      tmpNurbsL->makeEdge();
	      end = true;
	    } 
	  }while(!end); 	  
	}
	
	//*********************************************//  	
	//******* Search for 2D Elements: Lcmb ********//
	//*********************************************//
	if(line.find("LCMB")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16];

	  strLine>>cBuff;//Proceede to next word
	  strLine>>cBuff;//Proceede to next word
	  Lcmb * tmpLcmb = new Lcmb(cBuff);
	  elem2dVec.push_back(tmpLcmb);
	  strLine>>cBuff;
	  bool forward;
	  //cout<<"LCMB "<<tmpLcmb->getName()<<" : ";
	  while(strcmp(cBuff,"+")==0 || strcmp(cBuff,"-")==0)
	  {
	    if(strcmp(cBuff,"+")==0) forward=true;
	    else forward=false;
	    strLine>>cBuff;//cBuff contains now the Name of a 2d Element
	    for(vector<Element2d*>::iterator e = elem2dVec.begin();e != elem2dVec.end();++e)//Find 2d Element form list
	    {
	      if(strcmp(cBuff,(*e)->getName())==0)//Element found
	      {
		 tmpLcmb->addElement2d(*e,forward);
		 //cout<<" + "<< (*e)->getName();
	      }
	    }
	    strLine>>cBuff;
	  }
	  lcmbVec.push_back(tmpLcmb);
	  tmpLcmb->makeEdge();
	  //cout<<endl;
	}
	
	//*********************************************//  	
	//********** Search for Faces: GSUR ***********//
	//*********************************************//
	if(line.find("GSUR")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16],name[9];
	  bool blend,add;
	  strLine>>cBuff;//Proceede to next word
	  strLine>>cBuff;//Proceede to next word
	  strcpy(name,cBuff);	
	  Gsur * tmpGsur;
	  strLine>>cBuff;//Proceede to next word
	  if(strcmp(cBuff,"ADD")==0)
	  {
	    for(vector<Gsur*>::iterator s = gsurVec.begin();s != gsurVec.end();++s)
	    {	    
		if(strcmp(name,(*s)->getName())==0){
		  tmpGsur = (*s);break;}
	    }	
	    add=true;
	  }
	  else
	  {
	    tmpGsur = new Gsur(name);
	    gsurVec.push_back(tmpGsur);	  
	    strLine>>cBuff;//Proceede to next word
	    add=false;
	  }
  
	  if(strcmp(cBuff,"BLEND")==0) //BLEND Surface
	  {
	    //cout<<"Surface "<<tmpGsur->getName()<<" BLEND"<<endl;
	    blend = true;
	    tmpGsur->setBlend(true);
	  }
	  else //NURBS Surface
	  {
	    bool found = false;
	    for(vector<Surface*>::iterator n = nurbsVec.begin();n != nurbsVec.end();++n)//Find underlying Nurbs
	    {
		if(strcmp(cBuff,(*n)->getName())==0){
		  tmpGsur->setSurf((*n)->getSurf());
		  found = true;
		  break;
		}
	    }
	    if(!found && !add){
	      blend = true;
	      tmpGsur->setBlend(true);
	      cout<<"NURBS Surface "<<cBuff<<" not found ........................"<<endl;
	    }
	  }
	  
	  strLine>>cBuff;//Proceede to next word
        #if PLOT == 1
	  cout<<"Gsur "<<tmpGsur->getName()<<" : ";
	#endif
	  int edgeCounter = 0;
	  bool forward;
	  while(strcmp(cBuff,"+")==0 || strcmp(cBuff,"-")==0)//Search for Edges (2D Elements)
	  {
	    if(strcmp(cBuff,"+")==0) forward=true;
	    else forward=false;
	    strLine>>cBuff;//cBuff contains now the name of a 2d Element
	    for(vector<Element2d*>::iterator e = elem2dVec.begin();e != elem2dVec.end();++e)//Find 2d Element form list
	    {
	      if(strcmp(cBuff,(*e)->getName())==0)//Element found
	      {
		 tmpGsur->addWire((*e)->getWire(),forward,(*e)->getDiv());//Linkt Edge (Wire) to class Gsur
		 edgeCounter++;
		#if PLOT == 1
		 cout<<" + "<< (*e)->getName();	  
		#endif
		 break;
	      }
	    }
	    strLine>>cBuff;
	  }
	  if(edgeCounter == 0)
	    gsurVec.pop_back();
	 #if PLOT == 1
	  cout<<endl;
	 #endif
	}
	//*********************************************//  	
	//********* Search for Surface: NURS **********//
	//*********************************************//
	if(line.find("NURS")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16];	  
	  Surface * tmpNurbs;
	  bool end = false, compact = false, found, multi = false;
	  strLine>>cBuff;//Proceede to next word
	  strLine>>cBuff;//Proceede to next word	
	 #if PLOT == 1 
	  cout<<"Surface "<<cBuff<<"... NURBS"<<endl;
	 #endif
	  tmpNurbs = new Surface(cBuff);
	  strLine>>cBuff;//Proceede to next word
	  int uDeg,vDeg,nUKnt=0,nVKnt=0,nUPol,nVPol;	  
	  strLine>>cBuff;//Contains now uDeg or COMPACT
	  if(strcmp(cBuff,"COMPACT")==0)
	  {
	    compact=true,
	    strLine>>cBuff;//Contains now uDeg
	  }
	  uDeg=atoi(cBuff);
		  
	  strLine>>cBuff;//Contains now vDeg	
	  vDeg=atoi(cBuff);
	  
	  strLine>>cBuff;//Contains now nUpoles	
	  nUPol=atoi(cBuff);
	  
	  strLine>>cBuff;//Contains now nVpoles	
	  nVPol=atoi(cBuff);

	  TColgp_Array2OfPnt poles(1,nUPol,1,nVPol); 
	  TColStd_Array2OfReal weights(1,nUPol,1,nVPol); 

	  vector<double> uKtmp;
	  vector<double> vKtmp;
	  
	  vector<int> uMtmp;
	  vector<int> vMtmp;
	  
	  strLine>>cBuff;//Proceede to next word: DEFINE, CONTROL, KNOT or END
	  int initial=0, save=0;
	  do{    
	    //Save Line as vector of strings for easier access 
	    getline(input,line);
	    //Make line uppercase (Works only for ASCII Symbols)
	    for(int i=0;line[i]!=0;i++)
	    {
	      if(line[i]<=122 && line[i]>=97)
		line[i]-=32;
	    }
	    stringstream tmp;
	    tmp<<line;
	    vector<string> content;	    
	    while (tmp >> cBuff) {
	      content.push_back(cBuff);
	    }
	 
	    initial = save;//Save the position where the last control point was found
	    
	    if(strcmp(content[2].c_str(),"CONTROL")==0)
	    {
	      if(compact)
	      {
		gp_Pnt pnt(atof(content[5].c_str()),atof(content[6].c_str()),atof(content[7].c_str()));	
		poles.SetValue(atoi(content[3].c_str()),atoi(content[4].c_str()),pnt);
		weights.SetValue(atoi(content[3].c_str()),atoi(content[4].c_str()),atof(content[8].c_str()));
	      }
	      else
	      {
		found = false;
		if(hashing){
		  Point * tmpPnt = pointVec[pointGetIndex(content[5].c_str())-1];
		  gp_Pnt pnt(*((tmpPnt)->getX()),*((tmpPnt)->getY()),*((tmpPnt)->getZ()));
		  poles.SetValue(atoi(content[3].c_str()),atoi(content[4].c_str()),pnt);
		}
		else
		{
		  //Starting the search from 'initial' was implemented to speed up the search in case of Control Points written in consecutive lines
		  do{
		    int counter=initial;
		    for(vector<Point*>::iterator p = pointVec.begin()+initial;p != pointVec.end();++p)
		    {   
		      if(strcmp(content[5].c_str(),(*p)->getName())==0){
			gp_Pnt pnt(*((*p)->getX()),*((*p)->getY()),*((*p)->getZ()));	     
			poles.SetValue(atoi(content[3].c_str()),atoi(content[4].c_str()),pnt);
			found = true;
			break;
		      }  
		      counter++;
		      save++;
		    }
		    if (!found) save = 0;		  
		    initial = 0;//In case the Control Point was not written in a consecutive line, start searching again from the beginning
		  }while(found == false);	
		}		
		weights.SetValue(atoi(content[3].c_str()),atoi(content[4].c_str()),atof(content[6].c_str()));
	      }
	    }
	    else if(strcmp(content[2].c_str(),"KNOT")==0)
	    {
	      if(strcmp(content[3].c_str(),"U")==0)
	      {
		if(atoi(content[4].c_str())>1){
		    if(atof(content[5].c_str())==*(uKtmp.end()-1))
		      multi = true;
		    else 
		      multi = false;
		}
		else{
		  multi = false;
		}
		
		if(multi)//Not a new knot -> Increase Multiplicity
		{
		  (*(uMtmp.end()-1))++;
		}
		else//New knot
		{
		  uKtmp.push_back(atof(content[5].c_str()));
		  uMtmp.push_back(1);
		  nUKnt++;
		} 	
	      }
	      if(strcmp(content[3].c_str(),"V")==0)
	      {
		if(atoi(content[4].c_str())>1){
		    if(atof(content[5].c_str())==*(vKtmp.end()-1))
		      multi = true;
		    else 
		      multi = false;
		}
		else{
		  multi = false;
		}
		
		if(multi)//Not a new knot -> Increase Multiplicity 
		{
		  (*(vMtmp.end()-1))++;
		}
		else//New knot
		{
		  vKtmp.push_back(atof(content[5].c_str()));
		  vMtmp.push_back(1);
		  nVKnt++;
		} 
	      }	 
	    }
	    else if(strcmp(content[2].c_str(),"END")==0)
	    {  
	      TColStd_Array1OfReal uKnots(1,nUKnt);   
	      TColStd_Array1OfReal vKnots(1,nVKnt); 
	      TColStd_Array1OfInteger uMults(1,nUKnt);   
	      TColStd_Array1OfInteger vMults(1,nVKnt); 
	      int c=1;
	      for(vector<double>::iterator e = uKtmp.begin();e != uKtmp.end();++e)
	      {
		uKnots.SetValue(c,*e);
		uMults.SetValue(c,uMtmp[c-1]);
		c++;
	      }  
	      c=1;
	      for(vector<double>::iterator e = vKtmp.begin();e != vKtmp.end();++e)
	      {
		vKnots.SetValue(c,*e);
		vMults.SetValue(c,vMtmp[c-1]);
		c++;
	      }
	    #if PLOT == 1  
	      for(c=1;c<=nUKnt;c++)
	      {
		cout<<"Knots "<<uKnots.Value(c)<<" Mults: "<<uMults.Value(c)<<endl;
	      }
	      
	      for(c=1;c<=nVKnt;c++)
	      {
		cout<<"Knots "<<vKnots.Value(c)<<" Mults: "<<vMults.Value(c)<<endl;
	      }
	     #endif
	      Handle(Geom_BSplineSurface) aSurf = new Geom_BSplineSurface(poles,weights,uKnots,vKnots,uMults,vMults,uDeg,vDeg);
	      tmpNurbs->setSurf(aSurf);
	      nurbsVec.push_back(tmpNurbs);
	      end = true;
	    } 
	  }while(end == false); 
	}
	
	//*********************************************//
	//***** Search for elementary surfaces: SHPE **//
	//*********************************************//
	if(line.find("SHPE")<=4)//Shapes: Planes, Cylinders, Cones, Spheres, Torodial Surfaces
	{
	  Surface * tmpShpe;
	  stringstream tmp;
	  char cBuff[16], name[9];;
	  double vec1[3],vec2[3],vecN[3];
	  tmp<<line;
	  vector<string> content;	    
	  while (tmp >> cBuff) {
	    content.push_back(cBuff);
	  }  

	  strcpy(name,content[1].c_str());	  
	  tmpShpe = new Surface(name);
	 #if PLOT == 1  
	  cout<<"Surface "<<name<<"...";
	 #endif
	  if(strcmp(content[2].c_str(),"PLN")==0)//Plane
	  {
	   #if PLOT == 1   
	    cout<<"Plane"<<endl;
	   #endif 
	    int found = 0;
	    gp_Pnt p1,p2,p3;
	    
	    for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	    {
	      if(strcmp(content[3].c_str(),(*p)->getName())==0)
	      {
		p1.SetX(*((*p)->getX()));
		p1.SetY(*((*p)->getY()));
		p1.SetZ(*((*p)->getZ()));
		found++;
	      }  
	      if(strcmp(content[4].c_str(),(*p)->getName())==0)
	      {
		p2.SetX(*((*p)->getX()));
		p2.SetY(*((*p)->getY()));
		p2.SetZ(*((*p)->getZ()));
		found++;
	      }
	      if(strcmp(content[5].c_str(),(*p)->getName())==0)
	      {
		p3.SetX(*((*p)->getX()));
		p3.SetY(*((*p)->getY()));
		p3.SetZ(*((*p)->getZ()));
		found++;
	      }
	      if(found >=3) break;
	    }
	    gp_Vec v1(p1,p2);
	    gp_Vec v2(p1,p3);
	    gp_Dir dir1(v1);
	    gp_Dir dir2(v2);   
	    vec1[0]=dir1.X();
	    vec1[1]=dir1.Y();
	    vec1[2]=dir1.Z();
	    vec2[0]=dir2.X();
	    vec2[1]=dir2.Y();
	    vec2[2]=dir2.Z();
	    vl_prod(vec1,vec2,vecN);
	    gp_Vec normVec(vecN[0],vecN[1],vecN[2]);
	    gp_Dir normDir(normVec);
	    
	    Handle(Geom_Plane) myPlane= new Geom_Plane(p1,normDir);
	    tmpShpe->setSurf(myPlane);
	    nurbsVec.push_back(tmpShpe);   
	  }
	  else if(strcmp(content[2].c_str(),"CYL")==0)//Cylinder
	  {
	   #if PLOT == 1   
	    cout<<"Cylinder"<<endl;
	   #endif 
	    int found = 0;
	    double radius;
	    gp_Pnt p1,p2;
	    
	    for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	    {
	      if(strcmp(content[3].c_str(),(*p)->getName())==0)
	      {
		p1.SetX(*((*p)->getX()));
		p1.SetY(*((*p)->getY()));
		p1.SetZ(*((*p)->getZ()));
		found++;
	      }  
	      if(strcmp(content[4].c_str(),(*p)->getName())==0)
	      {
		p2.SetX(*((*p)->getX()));
		p2.SetY(*((*p)->getY()));
		p2.SetZ(*((*p)->getZ()));
		found++;
	      }
	      if(found >=2) break;
	    }
	    radius = atof(content[5].c_str());
	    gp_Vec v1(p1,p2);
	    gp_Dir normDir(v1);
	    gp_Ax3 axis(p1,normDir);
	    Handle(Geom_CylindricalSurface) myCylinder= new Geom_CylindricalSurface(axis,radius);
	    tmpShpe->setSurf( myCylinder);
	    nurbsVec.push_back(tmpShpe); 	    
	  }
	  else if(strcmp(content[2].c_str(),"SPH")==0)//Sphere
	  {
	   #if PLOT == 1   
	    cout<<"Sphere"<<endl;
	   #endif 
	    double radius;
	    gp_Pnt p1,p2;
	    
	    for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	    {
	      if(strcmp(content[3].c_str(),(*p)->getName())==0)
	      {
		p1.SetX(*((*p)->getX()));
		p1.SetY(*((*p)->getY()));
		p1.SetZ(*((*p)->getZ()));
		break;
	      }  
	    }
	    radius = atof(content[4].c_str());
	    p2.SetX(p1.X()-1);
	    p2.SetY(p1.Y()-1);
	    p2.SetZ(p1.Z()-1);
	    gp_Vec v1(p1,p2);
	    gp_Dir normDir(v1);
	    gp_Ax3 axis(p1,normDir);
	    Handle(Geom_SphericalSurface) mySphere= new Geom_SphericalSurface(axis,radius);
	    tmpShpe->setSurf(mySphere);
	    nurbsVec.push_back(tmpShpe); 		  
	  }
	  else if(strcmp(content[2].c_str(),"TOR")==0)//Torus
	  {
	   #if PLOT == 1  
	    cout<<"Torus"<<endl;
	   #endif 
	    double radius1, radius2;
	    int found = 0;
	    gp_Pnt p1,p2;
	    for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	    {
	      if(strcmp(content[3].c_str(),(*p)->getName())==0)
	      {
		p1.SetX(*((*p)->getX()));
		p1.SetY(*((*p)->getY()));
		p1.SetZ(*((*p)->getZ()));
		found++;
	      }  
	      if(strcmp(content[5].c_str(),(*p)->getName())==0)
	      {
		p2.SetX(*((*p)->getX()));
		p2.SetY(*((*p)->getY()));
		p2.SetZ(*((*p)->getZ()));
		found++;
	      }
	      if(found >=2) break;
	    }
	    radius1 = atof(content[4].c_str());
	    radius2 = atof(content[6].c_str());
	    gp_Vec v1(p1,p2);
	    gp_Dir normDir(v1);
	    gp_Ax3 axis(p1,normDir);
	    Handle(Geom_ToroidalSurface) myTorus= new Geom_ToroidalSurface(axis,radius1,radius2);
	    tmpShpe->setSurf(myTorus);
	    nurbsVec.push_back(tmpShpe); 		    
	  }
	  else if(strcmp(content[2].c_str(),"CON")==0)//Cone
	  {
	   #if PLOT == 1  
	    cout<<"Cone"<<endl;
	   #endif
	    int found = 0;
	    double radius1,radius2,height,angle;
	    gp_Pnt p1,p2;
	    
	    for(vector<Point*>::iterator p = pointVec.begin();p != pointVec.end();++p)
	    {
	      if(strcmp(content[3].c_str(),(*p)->getName())==0)
	      {
		p1.SetX(*((*p)->getX()));
		p1.SetY(*((*p)->getY()));
		p1.SetZ(*((*p)->getZ()));
		found++;
	      }  
	      if(strcmp(content[4].c_str(),(*p)->getName())==0)
	      {
		p2.SetX(*((*p)->getX()));
		p2.SetY(*((*p)->getY()));
		p2.SetZ(*((*p)->getZ()));
		found++;
	      }
	      if(found >=2) break;
	    }
	    radius1 = atof(content[5].c_str());
	    radius2 = atof(content[6].c_str());
	    height = p1.Distance(p2);
	    angle = atan((radius1-radius2)/height);

	    gp_Vec v1(p1,p2);
	    gp_Dir normDir(v1);
	    gp_Ax3 axis(p1,normDir);
	    if(radius1 - radius2 != 0){
	      Handle(Geom_ConicalSurface) myCone= new Geom_ConicalSurface(axis,angle,radius1);
	      tmpShpe->setSurf(myCone);
	    }
	    else{    
	      Handle(Geom_CylindricalSurface) myCyl= new Geom_CylindricalSurface(axis,radius1);
	      tmpShpe->setSurf(myCyl);
	    }    
	    nurbsVec.push_back(tmpShpe); 		      
	  }
	  else
	  {
	    cout<<"WARNING: unknown surface type!"<<endl;
	  }
	}
	
	//*********************************************//  	
	//********** Search for Bodies: Gbod **********//
	//*********************************************//
	if(line.find("GBOD")<=4)
	{
	  stringstream strLine;
	  strLine<<line;//Save line as stringstream
	  char cBuff[16],name[9];
	  strLine>>cBuff;//Proceede to next word
	  strLine>>cBuff;//Proceede to next word
	  strcpy(name,cBuff);	
	  Gbod * tmpGbod;
	  strLine>>cBuff;//Proceede to next word	
	  
	  if(strcmp(cBuff,"ADD")==0)//Additional line of Body definition
	  {
	    for(vector<Gbod*>::iterator b = gbodVec.begin();b != gbodVec.end();++b)
	    {	    
		if(strcmp(name,(*b)->getName())==0){
		  tmpGbod = (*b);break;}
	    }	    
	  }
	  else//New Body Found
	  {
	    tmpGbod = new Gbod(name);
	    gbodVec.push_back(tmpGbod);	  
	  }
	  
	  strLine>>cBuff;//Proceede to next word	
	  //cout<<"Gbod "<<tmpGbod->getName()<<" : ";
	  while(strcmp(cBuff,"+")==0 || strcmp(cBuff,"-")==0)//Search for Faces 
	  {
	    strLine>>cBuff;//cBuff contains now the name of a 2d Element
	    for(vector<Gsur*>::iterator e = gsurVec.begin();e != gsurVec.end();++e)//Find Faces form list
	    {
	      if(strcmp(cBuff,(*e)->getName())==0)//Element found
	      {
		 tmpGbod->addFace((*e)->getFace());//Linkt Edge (Wire) to class Gsur
		 (*e)->setFree(false);
		 //cout<<" + "<< (*e)->getName();
	      }
	    }
	    strLine>>cBuff;
	  }	  
	}
      }
      input.close();
      return true;
    }
    else
    {
      cout<<"There is no such file!"<<endl;
      return false;
    }
}



void setSplineTol(double *tol)
{
    cout<<"New Tolerance for Spline to NURBS conversion set: ";
    cout<<*tol<<endl;
    if(!usePia)
      cout<<"If NURBS lines start to oscillate increase allowed tollerance!"<<endl;
    ofstream tolContainer;
    tolContainer.open("Tolcontainer.txt",ios::out | ios::trunc);//Empty File
    tolContainer<<"Tol: "<<*tol<<endl;
    splineTolerance = *tol;
}

void getSplineTol()
{
    input.open("Tolcontainer.txt");
    string  content;    
    getline(input,content);
    if((content.substr(0,4)).compare("Tol:") != 0)//Default
    {
      cout<<"Default --> ";
      splineTolerance = 0.0001;
    }
    else 
    {
      splineTolerance = atof((content.substr(5,15)).c_str());
      cout<<"Loaded --> ";
    }
    if(splineTolerance<=0) splineTolerance = 0.0001;//Default
    cout<<"Current Tolerance: "<<splineTolerance<<endl;
    input.close();
}

//Save shape as STEP File "name"
void saveStep(TopoDS_Shape * shape,char * name)
{
  cout<<"Save start--------------------"<<endl;
  char myName[16];
  strcpy(myName,name);
  strcat (myName,".stp");
  if(!(*shape).IsNull())
  {
    STEPControl_Writer writer;
    //GeometricCurveSet
    if(writer.Transfer(*shape,STEPControl_AsIs)) cout<<"Saving "<<myName<<" successfully finished!"<<endl;
    writer.Write(myName); 
  }
  else
  {
    cout<<"Empty shape. Saving Failed!"<<endl;
  }
  cout<<"Save end----------------------"<<endl;
}

//Save shape as IGES File "name"
void saveIges(TopoDS_Shape * shape,char * name)
{
  cout<<"Save start--------------------"<<endl;
  char myName[16];
  strcpy(myName,name);
  strcat (myName,".igs");
  if(!(*shape).IsNull())
  {
    IGESControl_Writer writer("millimeters", 1);
    if(writer.AddShape(*shape)) cout<<"Saving "<<myName<<" successfully finished!"<<endl;
    writer.Write(myName); 
  }
  else
  {
    cout<<"Empty shape. Saving Failed!"<<endl;
  }
  cout<<"Save end----------------------"<<endl;
}

void startScreen()
{
  cout<<"======================================="<<endl;
  cout<<"*********  fbd2cad converter  *********"<<endl;
  cout<<"======================================="<<endl;  
}

void infoDisplay()
{
   cout<<"A fbd to STEP/IGES converter for the Finite Element Software CalculiX"<<endl;
   cout<<"Version 1.2 - 28 March 2017\n"<<endl;
   
   cout<<"by Pascal Mossier"<<'\n'<<endl;
   
   printf("usage: fbd2cad [parameter] filename \n\n");   
   printf("Parameter:\n");
   printf("Flags for export/import:\n");
   printf("\t '-wf'  -> export wireframe only\n"); 
   printf("\t '-bod' -> export bodies only (no free shells or free faces)\n");
   printf("\t '-hs'  -> use cgx alias names for entities to speed up parsing (works only if at least points use alias names)\n\n");
   printf("Flags for spline-fitting:\n");
   printf("\t '-tol' -> change Tolerance for spline-fitting (usage example: -tol 0.001) \n");
   printf("\t '-ofi' -> use Open CASCADE-algorithm for spline fitting. (might cause oscillation)\n");
   printf("\t '-sp'  -> To be used in combination with '-ofi'. If used the track points of splines are used for nurbs-line conversion instead of the potential node locations.\n\n"); 
   printf("Remarks:\n");
   printf("\t Upper- and lower chars are not distinguished. All chars are converted to upperKey.\n");
   printf("\t All '\%%' and '!' in entity names are replaced by a blank.\n");


   printf("The result is saved as result.stp and result.igs\n");
}

//*****************************************************//
//****  Process input Shape: perform shape-healing  ***//
//****  Try to assemble Shells from faces           ***//
//****  Try to assemble bodies from shells          ***//
//*****************************************************//
void processShell(TopoDS_Shape * input, TopoDS_Shape * output)
{
    cout<<"Fix Faces start----------------"<<endl;
    TopoDS_Shape ResultShape;
    Handle(ShapeFix_Shape) sfs = new ShapeFix_Shape; 
    sfs->Init (*input);
    sfs->Perform();   
    ResultShape = sfs->Shape(); 
    
    Handle(ShapeFix_Wireframe) SFWF = new ShapeFix_Wireframe(ResultShape);  

    SFWF->ModeDropSmallEdges() = Standard_False;  
    SFWF->FixSmallEdges(); 
    SFWF->FixWireGaps();  
    ResultShape = SFWF->Shape();    
    cout<<"Fix Faces end------------------"<<endl;
    TopExp_Explorer e;
    int counter = 0;
    //********* Build Shells **********//
    cout<<"Build Shells start-------------"<<endl;
    BRepBuilderAPI_Sewing shellMaker(shellSewingTolerance);
    
    for (e.Init(ResultShape,TopAbs_FACE);e.More(); e.Next())
    {
	counter++;
	shellMaker.Add(TopoDS::Face(e.Current()));
    }
    cout<<counter<<" faces processed"<<endl;
    shellMaker.Perform();
    TopoDS_Shape shell = shellMaker.SewedShape();
    
    counter = 0;
    for (e.Init(shell,TopAbs_SHELL);e.More(); e.Next()) counter++;
    cout<<counter<<" shells build"<<endl;
    if(counter==0) {return;}
    
    cout<<"Build Shells end---------------"<<endl; 
    
    if(shell.IsNull() || wireFrame){ 
      cout<<"No Faces Assembled --> Export only Wireframe"<<endl;
      *output = TestWire;
      return;
    }  
    cout<<"Fix Shell start----------------"<<endl;
    sfs = new ShapeFix_Shape; 
    sfs->Init (shell);
    sfs->Perform();   
    ResultShape = sfs->Shape(); 
    
    SFWF = new ShapeFix_Wireframe(ResultShape);  

    SFWF->ModeDropSmallEdges() = Standard_False;  
    SFWF->FixSmallEdges(); 
    SFWF->FixWireGaps();  
    ResultShape = SFWF->Shape(); 
    
    cout<<"Fix Shell end------------------"<<endl; 
    //********* Build Solids **********// 
    
    BRepBuilderAPI_MakeSolid solidmaker;
    TopTools_IndexedMapOfShape shellMap;
    TopExp::MapShapes(ResultShape, TopAbs_SHELL, shellMap);
    for(int ishell = 1; ishell <= shellMap.Extent(); ++ishell) {
	const TopoDS_Shell& shell = TopoDS::Shell(shellMap(ishell));
	solidmaker.Add(shell);
    }
    TopoDS_Shape solid = solidmaker.Solid();
    *output = solid;   
}

int main(int argc, char *argv[])
{
    //Default settings
    bool bod = false, loadTol = false;
    useSeqaPoints = false;
    usePia = true;
    wireFrame = hashing = false;
    startScreen();
    if(argc <= 1)
    {
      infoDisplay();
      return 0;
    }
    char outName [16] = "result"; 
    BRep_Builder myResultBuilder;
    TopoDS_Compound Result;
    myResultBuilder.MakeCompound(Result);
    
    WireFrameBuilder.MakeWire(TestWire);   
    
    char inName [32] = "-empty-"; 
    
    if(argc >1)
    {
      strcpy(inName,argv[argc-1]);
    
      for(int i=1;i<argc;i++)
      {
	if(strcmp(argv[i],"-tol") == 0)
	{
	  if(argc-1 == i) {cout<<"Invalid Input! Type: toCad -tol 0.01 FILE.fbd"<<endl; return 0;} 
	  double tol = atof(argv[i+1]);
	  loadTol = true;
	  setSplineTol(&tol);
	}
	if(strcmp(argv[i],"-bod") == 0){
	  bod = true;
	  cout<<"Export bodies only (no free shells or free faces)"<<endl;
	}
	if(strcmp(argv[i],"-ofi") == 0){
	  usePia = false;
	  cout<<"Use OCCT-algorithm for B-Spline fitting (might cause oscillation)"<<endl;
	}
	if(strcmp(argv[i],"-sp") == 0){
	  useSeqaPoints = true;
	  cout<<"Using Seqa points (no intpol3)"<<endl;
	}
	if(strcmp(argv[i],"-wf") == 0){
	  wireFrame = true;
	  cout<<"Exporting only the wireframe"<<endl;
	}
	if(strcmp(argv[i],"-hs") == 0){
	  hashing = true;
	  cout<<"Use cgx hashcode for faster reading"<<endl;
	}
      }
    }
    else
    {
      cout<<"Type 'fbd FILENAME.fbd'"<<endl;
      return 0;
    }
    
    if(!loadTol) getSplineTol();//load splineTolerance form container file
    
    cout<<"Parse fbd start----------------"<<endl;
    if(parseFBD(inName)) cout<<"-->Success"<<endl;
    else return 0;
    cout<<"Parse fbd end------------------"<<endl;
    
    cout<<"Assembling start---------------"<<endl;
    TopoDS_Shape resultFace;

    for(vector<Gsur*>::iterator f = gsurVec.begin();f != gsurVec.end();++f)
    {
    	(*f)->constructBoundary();
	(*f)->composeFace();  
    }
    int bodyCoutner = 0;
    cout<<"Assembling Bodies ---------------"<<endl;
    for(vector<Gbod*>::iterator b = gbodVec.begin();b != gbodVec.end();++b)
    {
      cout<<"Process "<<((*b)->getName())<<endl;
      TopoDS_Shape body;     
      BRep_Builder myBodyBuilder;
      processShell((*b)->joinFaces(&myBodyBuilder),&body); 
      if(!body.IsNull()) {
	myResultBuilder.Add(Result,body);
      bodyCoutner++;
      }
    }
    cout<<bodyCoutner<<" Bodies processed"<<endl;
    cout<<"Assembling Bodies end------------"<<endl;
   
    //Assemble and free Shells and Faces and add them to the comppound
    if(!bod){
      cout<<"Assembling Free Shells ----------"<<endl;
      TopoDS_Shell myShell;
      myBuilder.MakeShell(myShell);
      TopoDS_Shape resultFree;
      int freeCounter = 0;
      //Construct Faces
      for(vector<Gsur*>::iterator f = gsurVec.begin();f != gsurVec.end();++f)
      {
	  if((*f)->getFree() == true)//Surfaces that dont belong to a Body
	  {
	    myBuilder.Add(myShell,*((*f)->getFace()));
	    freeCounter++;
	  }
      }
      if(freeCounter>0){
	processShell(&myShell,&resultFree);
	if(resultFree.IsNull())
	  myResultBuilder.Add(Result,myShell);
	else
	  myResultBuilder.Add(Result,resultFree);      
      }
      else
      {
	if(wireFrame){
	  myResultBuilder.Add(Result,TestWire);
	}
      }
      cout<<"Assembling Free Shells end-------"<<endl;
    }
    cout<<"Assembling end-----------------"<<endl;
    
    
    //Write the resulting shape to a file
    saveStep(&Result,outName);
    saveIges(&Result,outName);
    return 0;
}

